/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "fProblem.h"
#include "fField.h"
#include "fBody.h"
#include "wTag.h"
#include "wElement.h"
#include "wMem.h"
#include "wCache.h"
using namespace tbox;
using namespace fpm;

Problem::Problem(std::shared_ptr<MshData> _msh, double aoa, double aos, double mch,
                 double sref, double cref, double xref, double yref, double zref) : msh(_msh), alpha(aoa), beta(aos), mach(mch), sR(sref), cR(cref)
{
    xR(0) = xref;
    xR(1) = yref;
    xR(2) = zref;
}

/**
 * @brief Set the fluid field
 */
void Problem::set(std::shared_ptr<Field> f)
{
    field = f;
}
/**
 * @brief Add a non-lifting body
 */
void Problem::add(std::shared_ptr<Body> b)
{
    bodies.push_back(b);
}

/**
 * @brief Compute velocity at element
 */
Eigen::Vector3d Problem::U(tbox::Element &e, std::vector<double> const &phi)
{
    Eigen::Vector3d v(0, 0, 0);
    // @todo to be implemented if usefull
    return v;
    std::cout << std::endl;
}

/**
 * @brief Compute density at element
 */
double Problem::Rho(tbox::Element &e, std::vector<double> const &phi)
{
    // particularized: pow(1 + 0.5 * (gamma - 1) * mInf * mInf * (1 - gradU2), 1 / (gamma - 1));
    Eigen::Vector3d u = U(e, phi);
    double a = 1 + 0.2 * mach * mach * (1 - u.dot(u));
    return sqrt(a * a * a * a * a);
}

/**
 * @brief Compute Mach number at element
 */
double Problem::Mach(tbox::Element &e, std::vector<double> const &phi)
{
    if (mach == 0)
        return 0;
    else
    {
        // particularized: pow(1 + 0.5 * (gamma - 1) * mInf * mInf * (1 - gradU2), 1 / (gamma - 1));
        Eigen::Vector3d u = U(e, phi);
        double a2 = 1 / (mach * mach) + 0.2 * (1 - u.dot(u)); // speed of sound squared
        return u.norm() / sqrt(a2);
    }
}

/**
 * @brief Compute pressure coefficient at element
 */
double Problem::Cp(tbox::Element &e, std::vector<double> const &phi)
{
    Eigen::Vector3d u = U(e, phi);
    if (mach == 0)
        return 1 - u.dot(u);
    else
    {
        //particularized: 2 / (gamma * mInf * mInf) * (pow(1 + 0.5 * (gamma - 1) * mInf * mInf * (1 - gradU2), gamma / (gamma - 1)) - 1);
        double a = 1 + 0.2 * mach * mach * (1 - u.dot(u));
        return 2 / (1.4 * mach * mach) * (sqrt(a * a * a * a * a * a * a) - 1);
    }
}

/**
 * @brief Update the elements memory (Jacobian)
 */
void Problem::updateMem()
{
    // Update volume Jacobian
    if (field)
        for (auto e : field->tag->elems)
            e->getVMem().update(true);
    // Update surface Jacobian
    for (auto s : bodies)
        for (auto e : s->tag->elems)
            e->getVMem().update(false);
}

/**
 * @brief Check that Problem is not empty and that elements are supported
 */
void Problem::check() const
{
    // Sanity checks
    if (bodies.empty())
        throw std::runtime_error("No lifting bodies provided!\n");
    if (mach != 0 && !field)
    {
        std::stringstream err;
        err << "Compressible computation requested for Mach "
            << mach << ", but no field provided!\n";
        throw std::runtime_error(err.str());
    }
    if (mach == 0 && field)
        std::cout << "Incompressible computation requested, but a field was provided. This will result in unnecessary computations!\n"
                  << std::endl;

    // Three-dimension problem
    // Volume elements
    if (field)
        for (auto e : field->tag->elems)
            if (e->type() != ELTYPE::HEX8)
            {
                std::stringstream err;
                err << "FPM solver is only implemented for surface elements of type "
                    << ELTYPE::HEX8 << " (" << e->type() << " was given)!\n";
                throw std::runtime_error(err.str());
            }
    // Surface elements
    for (auto b : bodies)
        for (auto e : b->tag->elems)
            if (e->type() != ELTYPE::QUAD4)
            {
                std::stringstream err;
                err << "FPM solver is only implemented for surface elements of type "
                    << ELTYPE::QUAD4 << " (" << e->type() << " was given)!\n";
                throw std::runtime_error(err.str());
            }
}

void Problem::write(std::ostream &out) const
{
    out << "fpm::Problem definition"
        << "\n\tAngle of attack: " << alpha
        << "\n\tAngle of sideslip: " << beta
        << "\n\tReference area: " << sR
        << "\n\tReference length: " << cR
        << "\n\tReference point: [" << xR.transpose() << "]"
        << std::endl;
    out << "with";
    for (auto b : bodies)
        out << "\t" << *b;
}
