/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#ifndef FPROBLEM_H
#define FPROBLEM_H

#include "fpm.h"
#include "wObject.h"
#include <memory>
#include <vector>
#include <Eigen/Dense>

namespace fpm
{

/**
 * @brief Manage the problem
 * @authors Adrien Crovato
 * @todo check and implement velocity computation
 */
class FPM_API Problem : public fwk::wSharedObject
{
public:
    std::shared_ptr<tbox::MshData> msh; ///< mesh data structure
#ifndef SWIG
    std::shared_ptr<Field> field;              ///< fluid field
    std::vector<std::shared_ptr<Body>> bodies; ///< bodies in fluid
#endif
    // Reference values
    double alpha;       ///< Angle of attack
    double beta;        ///< Angle of sideslip
    double mach;        ///< Mach number
    double sR;          ///< Reference surface
    double cR;          ///< Reference chord
    Eigen::Vector3d xR; ///< Reference center point (for moment computation)

    Problem(std::shared_ptr<tbox::MshData> _msh, double aoa, double aos, double mch, double sref, double cref, double xref, double yref, double zref);
    ~Problem() { std::cout << "~Problem()\n"; }

    void set(std::shared_ptr<Field> f);
    void add(std::shared_ptr<Body> b);

    // Functions
    inline double PhiI(Eigen::Vector3d const &pos);
    inline Eigen::Vector3d Ui();
    Eigen::Vector3d U(tbox::Element &e, std::vector<double> const &phi);
    double Rho(tbox::Element &e, std::vector<double> const &phi);
    double Mach(tbox::Element &e, std::vector<double> const &phi);
    double Cp(tbox::Element &e, std::vector<double> const &phi);

#ifndef SWIG
    void check() const;
    void updateMem();
    virtual void write(std::ostream &out) const override;
#endif
};

#include "fProblem.inl"

} // namespace fpm

#endif //FPROBLEM_H
