/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "fTimers.h"
#include <iomanip>

using namespace fpm;

Timer::Timer() : isOn(false)
{
    reset();
}

/**
 * @brief Reset the timer
 */
void Timer::reset()
{
    if (!isOn)
    {
        start();
        stop();
    }
    else
        throw std::runtime_error("Cannot reset a running timer!\n");
}

/**
 * @brief Start the timer
 */
void Timer::start()
{
    if (!isOn)
    {
        tS = std::chrono::high_resolution_clock::now();
        cS = std::clock();
        isOn = true;
    }
    else
        throw std::runtime_error("Cannot start a running timer!\n");
}

/**
 * @brief Stop the timer
 */
void Timer::stop()
{
    if (isOn)
    {
        tE = std::chrono::high_resolution_clock::now();
        cE = std::clock();
        isOn = false;
    }
    else
        throw std::runtime_error("Cannot stop a non-running timer!\n");
}

/**
 * @brief Access/create a given timer
 */
Timer &Timers::operator[](std::string const &name)
{
    return timers[name];
}

/**
 * @brief Print the times of all timers
 */
void Timers::write(std::ostream &out) const
{
    out << std::fixed << std::setprecision(1);
    out << std::setw(20) << std::left << "Timer"
        << std::setw(15) << std::right << "Wall-clock (s)"
        << std::setw(15) << std::right << "CPU (s)"
        << std::endl;

    for (auto const &tm : timers)
    {
        out << std::setw(20) << std::left << tm.first
            << std::setw(15) << std::right << tm.second.getWall()
            << std::setw(15) << std::right << tm.second.getCpu()
            << std::endl;
    }
}
